import logging
from pydantic import BaseModel, Field, PrivateAttr
from ._utils import JinjaVariables, render_jinja
from typing import Any, Dict, Union
from ._timeslice import Timeslice
import os
from ._stage_type import StageType
try:
    from databricks.sdk.runtime import * # noqa F403
except ModuleNotFoundError:
    pass


class DeltaLake(BaseModel):
    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._render()
        self.create_table()

    _logger: Any = PrivateAttr(default=None)
    _replacements: Dict[JinjaVariables, str] = PrivateAttr(default=None)
    database: str = Field(...)
    destination_table: str = Field(...)
    table: str = Field(...)
    container: str = Field(...)
    root: str = Field(...)
    path: str = Field(...)
    options: Union[dict, None] = Field(default=None)
    timeslice: Timeslice = Field(...)
    location: str = Field(default=None)
    checkpoint: str = Field(default=None)
    table_properties: Dict[str, str] = Field(default=None)
    stage: StageType = Field(...)

    def _render(self):
        self._replacements = {
            JinjaVariables.TABLE: self.destination_table,
            JinjaVariables.DATABASE: self.database,
            JinjaVariables.CONTAINER: self.container,
            JinjaVariables.CHECKPOINT: self.checkpoint,
        }

        self.root = render_jinja(self.root, self._replacements)
        self.path = render_jinja(self.path, self._replacements)
        self.database = render_jinja(self.database, self._replacements)
        self.table = render_jinja(self.table, self._replacements)
        if self.options:
            for option, value in self.options.items():
                self.options[option] = render_jinja(value, self._replacements)

        self.location = os.path.join(self.root, self.path)

    def create_table(self):
        table_ddl = f"""
            CREATE TABLE IF NOT EXISTS `{self.database}`.`{self.table}`
            USING DELTA
            LOCATION '{self.location}'
        """
        # add in the delta properties if there are any
        if self.table_properties:
            tbl_properties = [
                f"{k.lower()} = {v.lower()}" for k, v in self.table_properties.items()
            ]
            tbl_properties = ", ".join(tbl_properties)
            table_ddl = f"""{table_ddl}
            TBLPROPERTIES({tbl_properties})
            """
        try:
            spark.sql(f"CREATE DATABASE IF NOT EXISTS `{self.database}`")  # noqa F821
            spark.sql(table_ddl)  # noqa F821
        except NameError:
            version = os.getenv("DATABRICKS_RUNTIME_VERSION", None)
            if version:
                raise Exception("Spark undefined on databricks runtime!")
            logging.warning(
                "spark is not defined, skipping Spark operation for testing purposes"
            )



import logging
from pydantic import BaseModel, Field, PrivateAttr
from ._utils import JinjaVariables, render_jinja, get_ddl, load_schema
from typing import Any, Dict, List, Union
from ._timeslice import Timeslice
from enum import Enum
import os
from pyspark.sql.types import StructType
from pyspark.sql.streaming import StreamingQuery
from pyspark.sql import DataFrame
from ._stage_type import StageType


class TriggerType(Enum):
    File = "file"


class Source(BaseModel):
    _OPTION_CF_SCHEMA_HINTS = "cloudFiles.schemaHints"
    _OPTION_CORRUPT_RECORD_NAME = "columnNameOfCorruptRecord"

    def __init__(self, **data: Any) -> None:
        super().__init__(**data)
        self._logger = logging.getLogger(self.__class__.__name__)
        self._render()
        self.path = os.path.join(self.root, self.filename)

    _logger: Any = PrivateAttr(default=None)
    _replacements: Dict[JinjaVariables, str] = PrivateAttr(default=None)
    trigger: str = Field(default=None)
    trigger_type: TriggerType = Field(default=None)
    database: str = Field(...)
    destination_table: str = Field(...)
    table: str = Field(...)
    container: str = Field(...)
    root: str = Field(...)
    filename: str = Field(...)
    filename_date_format: str = Field(...)
    path_date_format: str = Field(...)
    options: dict = Field(...)
    timeslice: Timeslice = Field(...)
    format: str = Field(...)
    path: str = Field(default=None)
    spark_schema: Union[StructType, str] = Field(default=None)
    ddl: List[str] = Field(default=None)
    headerless_ddl: List[str] = Field(default=None)
    checkpoint: str = Field(default=None)
    stage: StageType = Field(...)

    def _render(self):
        self._replacements = {
            JinjaVariables.FILENAME_DATE_FORMAT: self.timeslice.strftime(
                self.filename_date_format
            ),
            JinjaVariables.PATH_DATE_FORMAT: self.timeslice.strftime(
                self.path_date_format
            ),
            JinjaVariables.TABLE: self.destination_table,
            JinjaVariables.DATABASE: self.database,
            JinjaVariables.CONTAINER: self.container,
            JinjaVariables.CHECKPOINT: self.checkpoint,
        }

        self.root = render_jinja(self.root, self._replacements)
        self.filename = render_jinja(self.filename, self._replacements)
        self.database = render_jinja(self.database, self._replacements)
        self.table = render_jinja(self.table, self._replacements)
        self.trigger = render_jinja(self.trigger, self._replacements)

        if self.options:
            for option, value in self.options.items():
                self.options[option] = render_jinja(value, self._replacements)

        self._config_schema_hints()

        if isinstance(self.spark_schema, str):
            path = self.spark_schema
            path = render_jinja(self.spark_schema, self._replacements)
            self._load_schema(path)

        corrupt_record_name = self.options.get(self._OPTION_CORRUPT_RECORD_NAME, None)
        if isinstance(self.spark_schema, StructType) and corrupt_record_name:
            if corrupt_record_name not in self.spark_schema.names:
                self.spark_schema.add(field=corrupt_record_name, data_type="string")

    def _config_schema_hints(self):
        path = self.options.get(self._OPTION_CF_SCHEMA_HINTS, None)
        if path and "/" in path:
            self._load_schema(path)

            if self.options.get("header"):
                self.options[self._OPTION_CF_SCHEMA_HINTS] = ", ".join(self.ddl)
            else:
                self.options[self._OPTION_CF_SCHEMA_HINTS] = ", ".join(
                    self.headerless_ddl
                )

    def _load_schema(self, path: str):
        if not self.spark_schema or isinstance(self.spark_schema, str):
            self.spark_schema = load_schema(path)
        if not self.ddl:
            self.ddl = get_ddl(self.spark_schema, header=True)
        if not self.headerless_ddl:
            self.headerless_ddl = get_ddl(self.spark_schema, header=False)

    def rename_headerless(self, df: Union[StreamingQuery, DataFrame]):
        columns = [c for c in df.columns if c not in ["_rescued_data"]]
        columns_cnt = len(columns)
        ddls = len(self.ddl)
        if columns_cnt != ddls:
            raise Exception(
                f"Headless files with schema hints must have a fully hinted schema since it must work positionally. Datasets!=dll({columns_cnt}!={ddls}"
            )

        for i, c in enumerate(columns):
            from_name = f"_c{i}"
            to_name = self.ddl[i].split(" ")[0].strip()
            logging.info(f"rename {from_name} to {to_name}")
            df: Union[StreamingQuery, DataFrame] = df.withColumnRenamed(
                from_name, to_name
            )

        return df

    class Config:
        arbitrary_types_allowed = True



import yaml
import os
from ._source import Source
from ._deltalake import DeltaLake
from ._timeslice import Timeslice
from typing import Union
from ._table_index import Tables, StageType, Table, _INDEX_WILDCARD


class Config:
    _CONFIG_PATH = "../Config/"
    _CONFIG_FILE = "config.yaml"
    _ENCODING = "utf-8"
    _BRONZE = "raw"
    _SILVER = "base"
    _SOURCE = "source"
    _TABLES = "tables"
    _TIMESLICE = "timeslice"
    _TABLE = "destination_table"
    _SOURCE_TABLE = "source_table"

    def __init__(self, config_path: str = None):
        self.config = {}

        if not config_path:
            config_path = os.path.join(self._CONFIG_PATH, self._CONFIG_FILE)

        with open(config_path, "r", encoding=self._ENCODING) as f:
            self.config = yaml.safe_load(f)

        _tables_path = self.config["tables"]

        with open(_tables_path, "r", encoding=self._ENCODING) as f:
            self.config["tables"] = yaml.safe_load(f)

        self.tables = Tables(table_data=self.config["tables"])

    def get_table_mapping(
        self,
        timeslice: Timeslice,
        stage: StageType,
        table: str = _INDEX_WILDCARD,
        database: str = _INDEX_WILDCARD,
        index: str = None,
    ):
        table_mapping = self.tables.get_table_mapping(
            stage=stage, table=table, database=database, index=index
        )

        if isinstance(table_mapping.source, dict):
            for name, table_obj in table_mapping.source.items():
                stage_config = self._get_stage_config(table_obj, timeslice)
                if table_obj.stage == StageType.landing:
                    table_mapping.source[name] = Source(**stage_config)
                if table_obj.stage != StageType.landing:
                    table_mapping.source[name] = DeltaLake(**stage_config)

        elif isinstance(table_mapping.source, Table):
            stage_config = self._get_stage_config(table_mapping.source, timeslice)
            if table_mapping.source.stage == StageType.landing:
                table_mapping.source = Source(**stage_config)
            if table_mapping.source.stage != StageType.landing:
                table_mapping.source = DeltaLake(**stage_config)

        stage_config = self._get_stage_config(table_mapping.destination, timeslice)
        if table_mapping.destination.stage in (StageType.base, StageType.raw):
            table_mapping.destination = DeltaLake(**stage_config)
        else:
            raise Exception(
                f"Table mapping destinations must be {StageType.raw} ro {StageType.base}"
            )

        return table_mapping

    def _get_stage_config(self, table: Table, timeslice: Timeslice):
        stage_config = self.config[table.stage.name]
        stage_config[self._TIMESLICE] = timeslice
        stage_config[self._TABLE] = table.name
        stage_config["table_properties"] = table.table_properties
        stage_config["stage"] = table.stage
        stage_config["database"] = table.database

        return stage_config

    def link_checkpoint(
        self,
        source: Union[Source, DeltaLake],
        destination: DeltaLake,
        checkpoint_name: str = None,
    ):
        if not checkpoint_name:
            checkpoint_name = f"{source.database}.{source.table}-{destination.database}.{destination.table}"

        source.checkpoint = checkpoint_name
        source._render()
        destination.checkpoint = checkpoint_name
        destination._render()
